﻿using System.Net.Sockets;
using NModbus;

namespace ModbusClientApplication
{
    class Program
    {
        // add here the list of IP addresses of one or more K-Ducer controllers
        public static List<string> list_of_KDU_IP_addresses = new List<string> { "192.168.32.103", "192.168.32.104", "192.168.32.105" };

        // change file path here if needed
        public static string fpath = "./";
        private static IModbusMaster? create_KDU_MB_client(string KDU_IP_Address)
        {
            try
            {
                // open TCP connection on port 502 (default modbus port)
                Console.WriteLine("Attempting to connect to KDU with IP Address " + (string)KDU_IP_Address);
                TcpClient client = new TcpClient( (string)KDU_IP_Address , 502 );
                client.SendTimeout = 500; // 500ms timeout for modbus messages
                Console.WriteLine("Successfully connected to KDU with IP Address " + (string)KDU_IP_Address);

                // create modbus client using NModbus library
                var factory = new ModbusFactory();
                IModbusMaster kdu_MB_client = factory.CreateMaster(client);
                return kdu_MB_client;
            }
            catch
            {
                Console.WriteLine("Unable to connect to KDU at IP Address: " + KDU_IP_Address);
                return null;
            }
        }

        // this function converts the big-endian byte array of the ASCII encoded KDU results string to a C# string
        // in python, this would be a one-liner: b''.join([i.to_bytes(2, byteorder='big') for i in printstring_in_modbus_format]).decode(encoding='ascii',errors='ignore')
        private static string? decode_result_printstring(ushort[]? printstring_in_modbus_format)
        {
            if(printstring_in_modbus_format == null)
                return null;

            byte[] reversed_byte_array = new byte[250];

            for(int i = 0; i < printstring_in_modbus_format.Length; i++)
            {
                byte[] two_chars_bigendian = BitConverter.GetBytes(printstring_in_modbus_format[i]);
                ushort big_endian_value = System.Buffers.Binary.BinaryPrimitives.ReadUInt16BigEndian(two_chars_bigendian);
                byte[] two_chars = BitConverter.GetBytes(big_endian_value);
                reversed_byte_array[i*2] = two_chars[0];
                reversed_byte_array[i*2 + 1] = two_chars[1];
            }
            return new string( System.Text.Encoding.ASCII.GetChars(reversed_byte_array) ).Trim(['\0','\r','\n']);
        }
        
        private static ushort[]? get_result_printstring(IModbusMaster kdu_MB_client)
        {
            ushort[] printstring_in_modbus_format = null!;
            try { printstring_in_modbus_format = kdu_MB_client.ReadInputRegisters( 0, (ushort)1-1, 125 ); }
            catch { ; } 
            return printstring_in_modbus_format;
        }

        private static ushort get_result_flag(IModbusMaster kdu_MB_client)
        {
            ushort[] result_flag = null!;
            try { result_flag = kdu_MB_client.ReadInputRegisters( 0, (ushort)295-1, 1 ); }
            catch { return 2; } // special value for error detection
            // the function above has three parameters: slaveAddress (ignored by the K-Ducer), startAddress (the holding register address to start reading from), and numberOfPoints which is the number of consecutive registers to read
            if( result_flag == null )
                return 2; // special value for error detection
            return result_flag[0];
            /* according to the K-Ducer Modbus Map, input register 295:
                    Self-clearing flag indicating if a new result is available.
                    0 = No new screwdriving result
                    1 = New screwdriving result available
                    Every time this register is read, its value is reset to 0. Every time there is a new result, this value is set to 1.
                    This value is meant to be polled cyclically to check when new results data is available.
            */
        }
        
        private static string get_station_name(IModbusMaster kdu_MB_client)
        {
            string? station_name = null;
            try { station_name = decode_result_printstring(kdu_MB_client.ReadInputRegisters( 0, (ushort)381-1, 13 )); }
            catch { ; }
            if(station_name == null)
                return "";
            return station_name;
        }

        private static void continuously_poll_KDU_for_results(Object? KDU_IP_Address)
        {
            if( KDU_IP_Address == null)
                return;
            
            // create client and connect
            IModbusMaster? kdu_MB_client = create_KDU_MB_client((string)KDU_IP_Address);
            string fname = "";
            // error timer to prevent spam
            System.Diagnostics.Stopwatch errTimer = System.Diagnostics.Stopwatch.StartNew();
            while( true )
            {
                Thread.Sleep( 50 ); // 50ms pause between requests
                // if not connected, attempt to reconnect
                if(kdu_MB_client == null)
                {
                    kdu_MB_client = create_KDU_MB_client((string)KDU_IP_Address);
                    continue;
                }
                else if(fname == "")
                {
                    // set file name
                    fname = get_station_name(kdu_MB_client) + " - " + (string)KDU_IP_Address + " - " + DateTime.Now.ToString("yyyy-MM-dd") + ".txt";
                    // add header if file is empty
                    try
                    {
                        if (!File.Exists(fpath+fname) || new FileInfo(fpath+fname).Length == 0 )
                        {
                            File.AppendAllText(fpath+fname, "\n");
                        }
                    }
                    catch { Console.WriteLine("Error creating file " + fpath+fname); }
                    continue;
                }

                ushort result_flag = get_result_flag(kdu_MB_client);
                if( result_flag == 2 ) // error
                {
                    if( errTimer.ElapsedMilliseconds < 5000 ) // print only every 5 secondds
                        continue; // skips timer restart
                    else
                    {
                        Console.WriteLine("Error retrieving result for KDU at IP Address " + (string)KDU_IP_Address);
                        try { kdu_MB_client.Dispose(); }
                        finally { kdu_MB_client = null; }
                    }
                }
                else if( result_flag == 1) // has new result
                {
                    string? kducer_result = decode_result_printstring(get_result_printstring(kdu_MB_client));
                    if(kducer_result == null || graph_data == null)
                    {
                        try { kdu_MB_client.Dispose(); }
                        finally { kdu_MB_client = null; }
                        continue;
                    }

                    try
                    {
                        File.AppendAllText(fpath+fname, kducer_result + "\n");
                        Console.WriteLine("New result saved to " + fpath+fname + ", result: " + kducer_result);
                    }
                    catch { Console.WriteLine("Error saving to " + fpath+fname + "\n    discarding result: " + kducer_result); }
                }
                errTimer.Restart();
            }
        }
        
        public static void Main(string[] args)
        {
            foreach (var KDU_IP_Address in list_of_KDU_IP_addresses)
            {
                Thread KDU_Poller_Thread = new Thread(continuously_poll_KDU_for_results);
                KDU_Poller_Thread.Start(KDU_IP_Address);
                Thread.Sleep( 50 / list_of_KDU_IP_addresses.Count );
            }

            while(true)
            {
                Thread.Sleep(0);
            }
        }
    }
}
