#example of retreiving screwdriving results via MODBUS remote communication with K-Ducer
#uses python3 with pymodbustcp library from https://pypi.org/project/pyModbusTCP/
#install with pip install pyModbusTCP

# note: this is just an example to illustrate some basic functionality
# Kolver offers K-Link as a free and more advanced solution to record the tightening results over a network

from pyModbusTCP.client import ModbusClient
from datetime import date

address_offset = -1 #per modbus specifications: MODBUS addresses are indexed at 1, but in the MODBUS PDU, they are indexed at 0
#some MODBUS implementations, for example Siemens Step7 PLCs, automatically add this offset

#the list of IP addresses of one or more K-Ducer controllers:
kdu_ip_address_list = [ '192.168.32.103' , '192.168.32.104' , '192.168.32.105' ]
#the file path where to save the results. default './' to save in the same directory of this script.
results_file_directory = ''

def decode_kducer_printstring(output_from_function_read_input_registers_0_125):
    #input: list of integers exactly as received from K-DUCER response to read_input_registers(reg_addr=0, reg_nb=125)
    #returns: string containing screwdriving results
    #this is necessary because pyModbusTCP assumes all responses are integers, but the K-DUCER response is an ASCII coded string
    #must convert each integer to two bytes in big endian (modbus uses big endian), then decode as ASCII
    if output_from_function_read_input_registers_0_125 == None or type(output_from_function_read_input_registers_0_125) != type(['']) or len(output_from_function_read_input_registers_0_125) == 0:
        return ''
    return b''.join([i.to_bytes(2, byteorder='big') for i in output_from_function_read_input_registers_0_125]).decode(encoding='ascii',errors='ignore').replace('\00','').replace('\r\n','')

def check_new_result_flag(KDU_modbus_client):
    result_flag = KDU_modbus_client.read_input_registers(reg_addr= address_offset+295, reg_nb= 1)
    if result_flag and result_flag[0] == 1:
        return True
    else:
        return False

def get_result_printstring(KDU_modbus_client):
    screwdriving_results_raw_modbus = KDU_modbus_client.read_input_registers(reg_addr= address_offset+1, reg_nb= 125)
    screwdriving_results_decoded = decode_kducer_printstring(screwdriving_results_raw_modbus)
    return screwdriving_results_decoded

def get_station_name(KDU_modbus_client):
    try:
        station_name_raw_modbus = KDU_modbus_client.read_input_registers(reg_addr= address_offset+381, reg_nb= 13)
        station_name_decoded = decode_kducer_printstring(station_name_raw_modbus)
        return station_name_decoded
    except:
        return ''

def print_graph_data(KDU_modbus_client):
    # note: these registers contain torque-angle graph data compressed to at most 70 data points
    # the full-resolution graph data is stored on the USB drive connected to the controller, if any, or via K-Link or via K-Graph
    TT_graph_data = KDU_modbus_client.read_input_registers(reg_addr= address_offset+153, reg_nb= 71)
    TA_graph_data = KDU_modbus_client.read_input_registers(reg_addr= address_offset+224, reg_nb= 71)
    print(f'Angle-time data: {TA_graph_data}')
    print(f'Torque-time data: {TT_graph_data}')

def save_result_to_file(result_string, filepath):
    with open(filepath, 'a') as f:
        f.write(result_string + '\n')

if __name__ == "__main__":
    # instantiate modbus clients
    kdu_modbus_client_list = []
    for kdu_ip_address in kdu_ip_address_list:
        KDU_MBClient = ModbusClient(host=kdu_ip_address, port=502, auto_open=True, auto_close=False, timeout=0.05) #instantiate a MODBUS client. auto_close must be False for REMOTE_LEVER control
        KDU_MBClient.debug = False #use True if you want the script to print all the MODBUS packets and debug informantion on the console
        kdu_modbus_client_list.append(KDU_MBClient)
    
    # get station names and generate file names
    result_file_path_list = []
    for kdu_MB_client in kdu_modbus_client_list:
        fpath = f'{results_file_directory}{get_station_name(kdu_MB_client)} - {kdu_MB_client.host} - {date.today()}.txt'
        result_file_path_list.append(fpath)

    # start polling for results...
    print(f'Now polling KDU controllers at IP addresses: {kdu_ip_address_list}')
    while 1:
        for idx, kdu_MB_client in enumerate(kdu_modbus_client_list):
            has_new_result = check_new_result_flag(kdu_MB_client)
            if has_new_result:
                result_string = get_result_printstring(kdu_MB_client)
                print(f'New result from {kdu_MB_client.host}: {result_string}')
                save_result_to_file(result_string, result_file_path_list[idx])
                print_graph_data(kdu_MB_client)