#example of MODBUS remote control with K-Ducer
#uses python3 with pymodbustcp library from https://pypi.org/project/pyModbusTCP/
#install with pip install pyModbusTCP

from pyModbusTCP.client import ModbusClient
import time

address_offset = -1 #per modbus specifications: MODBUS addresses are indexed at 1, but in the MODBUS PDU, they are indexed at 0
#some MODBUS implementations, for example Siemens Step7 PLCs, automatically add this offset
kdu_ip_address = '192.168.32.103' #the IP address of the K-Ducer unit. Set via Menu > General Settings > Communication Protocol

def read_screwdriving_state(MBclient):
    return MBclient.read_input_registers(reg_addr= address_offset+138, reg_nb= 1)[0] #read input register address 138: screwdriving state

def is_screwdriving_done(MBclient):
    # read input register address 295: new screwdriving result
    # this register is self-resetting, and will have a value of "1" only once per new screwdriving result
    return MBclient.read_input_registers(reg_addr= address_offset+295, reg_nb= 1)[0]

def change_current_program(MBclient, program_number):
    while True:
        MB_request_accepted = MBclient.write_single_register(reg_addr= address_offset+7373, reg_value= program_number) #write to holding register address 7373 to change current selected program
        if MB_request_accepted == True:
            print('Change program request accepted.')
            time.sleep(0.5) #wait 500ms to allow new program to load
            return True
        else:
            print('Change program request refused. Trying again in 3 seconds...')
            time.sleep(3)

def run_motor_until_it_stops(MBclient):
    #"REMOTE LEVER" MODBUS request must be sent at least every 100ms to maintain the screwdriver running. This a safety feature ensuring a fail-to-off system.
    while True:
        MBclient.write_single_coil(bit_addr= address_offset+33, bit_value= 1) #write 1 to coil address 33: "REMOTE LEVER" to run the screwdriver.
        if is_screwdriving_done(MBclient):
            MBclient.write_single_coil(bit_addr= address_offset+33, bit_value= 0) # reset the lever command. the lever bit also automatically resets after 500ms of inactivity
            break
    screwdriving_state = read_screwdriving_state(MBclient)
    return screwdriving_state

def print_closing_torque_and_angle(MBclient):
    last_closing_torque = MBclient.read_input_registers(reg_addr= address_offset+149, reg_nb= 2) #read input register 149: torque. Note the register span for torque is 2
    last_closing_torque_Nm = last_closing_torque[1]/100
    ''' as described in the MODBUS map, the torque value is split into two registers
        first register contains the thousandths of newton multiplied by 1000
        the thousandths of newtown are always zero for KDS series screwdriver size PL-6 and above
        second register contains the hundrethds of newton '''
    last_closing_angle = MBclient.read_input_registers(reg_addr= address_offset+151, reg_nb= 2)[1] #read input register 151: angle. Note the register span for angle is 2
    print('Last closing torque: {} Nm. Last closing angle: {} deg'.format(last_closing_torque_Nm,last_closing_angle))

if __name__ == "__main__":
    print('******WARNING******* screwdriver motor will start running in 10 seconds. Press CTRL+c to abort script.')
    time.sleep(1)
    exampleMBClient = ModbusClient(host=kdu_ip_address, port=502, auto_open=True, auto_close=False, timeout=3) #instantiate a MODBUS client. auto_close must be False for REMOTE_LEVER control
    exampleMBClient.debug = False #use True if you want the script to print all the MODBUS packets and debug informantion on the console
    
    #run screwdriver. If Screw OK, repeat with program #1. If Screw NOK, repeat with program #2. 
    next_program = 1
    while True:
        change_current_program(MBclient=exampleMBClient, program_number=next_program)
        final_screwdriving_state = run_motor_until_it_stops(MBclient=exampleMBClient)
        if (final_screwdriving_state == 13) or (final_screwdriving_state == 14): # 13 = Screw OK (torque control mode), 14 = Angle OK (angle control mode). Refer to K-Ducer MODBUS map.
            next_program = 1
            print('Screw OK. Selecting program 1')
        else:
            next_program = 2
            print('Screw N-OK. Selecting program 2')
        print_closing_torque_and_angle(MBclient=exampleMBClient)