#uses python3 with pymodbustcp library from https://pypi.org/project/pyModbusTCP/
#install with pip install pyModbusTCP

from pyModbusTCP.client import ModbusClient
import time

address_offset = -1 #per modbus specifications: MODBUS addresses are indexed at 1, but in the MODBUS PDU, they are indexed at 0
#some MODBUS implementations, for example Siemens Step7 PLCs, automatically add this offset
kdu_ip_address = '192.168.32.103' #the IP address of the K-Ducer unit. Set via Menu > General Settings > Communication Protocol

def reprogram_from_modbus_permanent_memory():
    # before v38, this was the only way to modify parameters in the KDU controller
    # first, we enter "remote programming mode"
    # then, we write the desired changes
    # then, we confirm the changes by exiting "remote programming mode"
    # this saves the changes in permanent memory, so if the controller is turned OFF then ON, the changes are retained
    # ------
    # in many situations where remote programming is used, it is not necessary to retain the changes after a power OFF-ON cycle
    # because the controlling device (PC, PLC, robot, etc) will always reprogram the KDU controller after a power cycle or before working
    # therefore we have introduced a new fast programming mode in v38 that only affects volatile memory
    # see reprogram_from_modbus_new_recommended_method() for details
    exampleMBClient = ModbusClient(host=kdu_ip_address, port=502, auto_open=True, auto_close=False, timeout=1) #instantiate a MODBUS client. Always use auto_close=False
    exampleMBClient.debug = True #use True if you want the script to print all the MODBUS packets and debug informantion on the console

    # enter programming mode
    exampleMBClient.write_single_register(reg_addr= address_offset+7790, reg_value= 1) # 7790 => 1

    # base address (modbus addresses for program number 1)
    addr_target_torque_Ncm = 2
    addr_max_torque_Ncm = 4
    addr_final_speed = 16
    addr_conf = 23 # 16 bits for various on/off options, see Modbus Map
    addr_number_of_screws = 68

    # modify some parameters for program 1
    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_target_torque_Ncm ), reg_value= 150 ) # torque target => 150 N-cm. Single register.

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_max_torque_Ncm ), reg_value= 180 ) # torque max => 180 N-cm. Single register.

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_final_speed ), reg_value= 300 ) # final speed => 300 RPM

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_conf ), reg_value= 1024) # Conf => Lever error ON (bit 10 counting from the right), everything else OFF => binary 0000 0100 0000 0000 (10th bit from right is 1) => hex 400 => decimal 1024
    
    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_number_of_screws ), reg_value= 12 ) # number of screws => 12

    # now modify the same parameters but for program 10
    nr_pr = 10

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_target_torque_Ncm + (115 * (nr_pr - 1)) ), reg_value= 150) # torque target => 150 N-cm

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_max_torque_Ncm ), reg_value= 180 ) # torque max => 180 N-cm. Single register.

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_final_speed + (115 * (nr_pr - 1)) ), reg_value= 300 ) # final speed => 300 RPM

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_conf + (115 * (nr_pr - 1)) ), reg_value= 1024) # Conf => Lever error ON (bit 10 counting from the right), everything else OFF => binary 0000 0100 0000 0000 (10th bit from right is 1) => hex 400 => decimal 1024
    
    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_number_of_screws + (115 * (nr_pr - 1)) ), reg_value= 12 ) # number of screws => 12
    
    # exit remote programming mode with value 1 to save changes
    exampleMBClient.write_single_register(reg_addr= address_offset+7790, reg_value= 2) # 7790 => 1
    time.sleep(5) # wait 5 seconds for system to reload programs

    # select program #10 (this will only work if the REMOTE PR general setting/option is set to CN5 TCP, refer to the k-ducer operator manual if in doubt)
    exampleMBClient.write_single_register(reg_addr=( address_offset + 7373 ), reg_value= 10) # switch to program #10

def reprogram_from_modbus_new_recommended_method():
    # in many situations where remote programming is used, it is not necessary to retain the changes after a power OFF-ON cycle
    # because the controlling device (PC, PLC, robot, etc) will always reprogram the KDU controller after a power cycle or before working
    # therefore we have introduced a new fast programming mode in v38 that only affects volatile memory
    # this is the recommended method as it is much faster to apply changes
    exampleMBClient = ModbusClient(host=kdu_ip_address, port=502, auto_open=True, auto_close=False, timeout=1) #instantiate a MODBUS client. Always use auto_close=False
    exampleMBClient.debug = True #use True if you want the script to print all the MODBUS packets and debug informantion on the console

    # base address (modbus addresses for program number 1)
    addr_target_torque_Ncm = 7792
    addr_max_torque_Ncm = 7794
    addr_final_speed = 7806
    addr_conf = 7813 # 16 bits for various on/off options, see Modbus Map
    addr_number_of_screws = 7858

    # modify some parameters for program 1
    # first, select program #1
    exampleMBClient.write_single_register(reg_addr=( address_offset + 7373 ), reg_value= 1)

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_target_torque_Ncm ), reg_value= 150 ) # torque target => 150 N-cm. Single register.

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_max_torque_Ncm ), reg_value= 180 ) # torque max => 180 N-cm. Single register.

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_final_speed ), reg_value= 300 ) # final speed => 300 RPM

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_conf ), reg_value= 1024) # Conf => Lever error ON (bit 10 counting from the right), everything else OFF => binary 0000 0100 0000 0000 (10th bit from right is 1) => hex 400 => decimal 1024
    
    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_number_of_screws ), reg_value= 12 ) # number of screws => 12

    # now modify the same parameters but for program 10
    # first, select program #10
    exampleMBClient.write_single_register(reg_addr=( address_offset + 7373 ), reg_value= 10)

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_target_torque_Ncm ), reg_value= 150 ) # torque target => 150 N-cm. Single register.

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_max_torque_Ncm ), reg_value= 180 ) # torque max => 180 N-cm. Single register.

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_final_speed ), reg_value= 300 ) # final speed => 300 RPM

    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_conf ), reg_value= 1024) # Conf => Lever error ON (bit 10 counting from the right), everything else OFF => binary 0000 0100 0000 0000 (10th bit from right is 1) => hex 400 => decimal 1024
    
    exampleMBClient.write_single_register(reg_addr=( address_offset + addr_number_of_screws ), reg_value= 12 ) # number of screws => 12

    # within ~1/4 of a second, the changes will become effective on the KDU
    # entering the main menu or powering off the KDU will discard the changes and reload the configuration from permanent memory

if __name__ == "__main__":
    # reprogram_from_modbus_permanent_memory() # use this function only if:
    #           - you have KDU software v37 or earlier (note: software updates are free!)
    #           - if you must retain the new parameters after powering the KDU controller OFF
    reprogram_from_modbus_new_recommended_method()



