#example of retreiving screwdriving results via MODBUS remote communication with K-Ducer
#uses python3 with pymodbustcp library from https://pypi.org/project/pyModbusTCP/
#install with pip install pyModbusTCP
#also uses numpy and matplotlib to graph results, install with pip install numpy and pip install matplotlib

from pyModbusTCP.client import ModbusClient
import time
import numpy as np
import matplotlib.pyplot as plt
import multiprocessing

address_offset = -1 #per modbus specifications: MODBUS addresses are indexed at 1, but in the MODBUS PDU, they are indexed at 0
#some MODBUS implementations, for example Siemens Step7 PLCs, automatically add this offset
kdu_ip_address = '192.168.32.103' #the IP address of the K-Ducer unit. Set via Menu > General Settings > Communication Protocol

def convert_kducer_printstring(output_from_function_read_input_registers_0_125):
    #input: list of integers exactly as received from K-DUCER response to read_input_registers(reg_addr=0, reg_nb=125)
    #returns: string containing screwdriving results
    #this is necessary because pyModbusTCP assumes all responses are integers, but the K-DUCER response is an ASCII coded string
    #must convert each integer to two bytes in big endian (modbus uses big endian), then decode as ASCII
    if output_from_function_read_input_registers_0_125 == None or type(output_from_function_read_input_registers_0_125) != type(['']) or len(output_from_function_read_input_registers_0_125) == 0:
        print('Input must be the list returned by the pyModbusTCP function: read_input_registers(reg_addr=0, reg_nb=125). Input was: {}'.format(output_from_function_read_input_registers_0_125))
        return ''
    return (b''.join([i.to_bytes(2, byteorder='big') for i in output_from_function_read_input_registers_0_125]).decode(encoding='ascii',errors='ignore')).replace('\00','').replace('\r\n','\n')

def get_graph_data(MBclient):
    time.sleep(0.25)
    TT_graph_data = MBclient.read_input_registers(reg_addr= address_offset+153, reg_nb= 71)
    TA_graph_data = MBclient.read_input_registers(reg_addr= address_offset+224, reg_nb= 71)
    TT_graph_data.extend(TA_graph_data)
    return TT_graph_data
    
def graph_results(graph_data, print_string):

    TT_sample_time = graph_data[0]
    TT_data_y = np.trim_zeros(graph_data[1:70], 'b')

    TA_sample_time = graph_data[71]
    TA_data_y = np.trim_zeros(graph_data[72:], 'b')

    #Screw NOK results can have unequal length of TA/TT results
    min_length = min([len(TT_data_y), len(TA_data_y)])
    TT_data_y = TT_data_y[:min_length]
    TA_data_y = TA_data_y[:min_length]

    TT_data_x = range(0,len(TT_data_y)*TT_sample_time,TT_sample_time)
    TA_data_x = range(0,len(TA_data_y)*TA_sample_time,TA_sample_time)

    plt.figure(clear=True, figsize=[14,6], dpi = 100, tight_layout=True)

    plt.subplot(131, title='Torque vs Time')
    plt.plot(TT_data_x, TT_data_y, marker='.')
    plt.grid()
    plt.ylabel('Torque, cNm')
    plt.xlabel('Time, ms')

    plt.subplot(132, title='Torque vs Angle')
    plt.plot(TA_data_y, TT_data_y, marker='.')
    plt.grid()
    plt.ylabel('Torque, cNm')
    plt.xlabel('Angle, degrees')

    plt.subplot(133, title='Angle vs Time')
    plt.plot(TA_data_x, TA_data_y, marker='.')
    plt.grid()
    plt.ylabel('Angle, degrees')
    plt.xlabel('Time, ms')

    plt.suptitle('K-Ducer Torque/Angle data from last screwdriving, obtained via MODBUS TCP\n\n{}'.format(print_string))
    plt.show()

def continously_check_screwdriving_results():
    exampleMBClient = ModbusClient(host=kdu_ip_address, port=502, auto_open=True, auto_close=False, timeout=3) #instantiate a MODBUS client. auto_close must be False for REMOTE_LEVER control
    exampleMBClient.debug(False) #use True if you want the script to print all the MODBUS packets and debug informantion on the console
    
    old_response = exampleMBClient.read_input_registers(reg_addr= address_offset+1, reg_nb= 125)
    screwdriving_results_string = convert_kducer_printstring(old_response)
    #print('Last result from K-Ducer with IP {}:\n{}'.format(exampleMBClient.host(),screwdriving_results_string))

    graph_in_another_core = None
    while 1:
        time.sleep(0.2) #check for a new result every 200ms. Place this loop in a thread to perform other tasks asynchronously
        new_response = exampleMBClient.read_input_registers(reg_addr= address_offset+1, reg_nb= 125)
        if new_response == old_response:
            continue #not a new result, skip rest of loop
        if graph_in_another_core is not None:
            graph_in_another_core.kill()
        old_response = new_response
        screwdriving_results_string = convert_kducer_printstring(new_response)
        print('New result from K-Ducer with IP {}:\n{}'.format(exampleMBClient.host(),screwdriving_results_string))
        graph_data = get_graph_data(exampleMBClient)
        graph_in_another_core = multiprocessing.Process(target=graph_results,args=(graph_data,screwdriving_results_string))
        graph_in_another_core.start()

if __name__ == "__main__":
    print('This program polls the K-Ducer for the latest screwdriving results and graphs the torque/angle data.\nPolling for new screwdriving results... Run the screwdriver to see the results.')
    continously_check_screwdriving_results()