#example of MODBUS remote control with K-Ducer
#uses python3 with pymodbustcp library from https://pypi.org/project/pyModbusTCP/
#install with pip install pyModbusTCP
#also uses numpy and matplotlib to graph results, install with pip install numpy and pip install matplotlib

from pyModbusTCP.client import ModbusClient
import time
import numpy as np
import matplotlib.pyplot as plt

address_offset = -1 #per modbus specifications: MODBUS addresses are indexed at 1, but in the MODBUS PDU, they are indexed at 0
#some MODBUS implementations, for example Siemens Step7 PLCs, automatically add this offset
kdu_ip_address = '192.168.32.103' #the IP address of the K-Ducer unit. Set via Menu > General Settings > Communication Protocol

def read_screwdriving_state(MBclient):
    return MBclient.read_input_registers(reg_addr= address_offset+138, reg_nb= 1)[0] #read input register address 138: screwdriving state

def change_current_program(MBclient, program_number):
    while True:
        MB_request_accepted = MBclient.write_single_register(reg_addr= address_offset+7373, reg_value= program_number) #write to holding register address 7373 to change current selected program
        if MB_request_accepted == True:
            print('Change program request accepted.')
            time.sleep(0.5) #wait 500ms to allow new program to load
            return True
        else:
            print('Change program request refused. Trying again in 3 seconds...')
            time.sleep(3)

def run_motor_until_it_stops(MBclient):
    #"REMOTE LEVER" MODBUS request must be sent at least every 100ms to maintain the screwdriver running. This a safety feature ensuring a fail-to-off system.
    while True:
        MBclient.write_single_coil(bit_addr= address_offset+33, bit_value= 1) #write 1 to coil address 33: "REMOTE LEVER" to run the screwdriver.
        time.sleep(0.1) #wait 100ms. This avoids excess network traffic and allows motor time to physically start before checking state
        screwdriving_state = read_screwdriving_state(MBclient)
        if screwdriving_state != 11: # 11 = tightening. Refer to K-Ducer MODBUS map.
            time.sleep(0.5) #wait 500ms to ensure system goes idle before proceeding
            return screwdriving_state

def convert_kducer_printstring(output_from_function_read_input_registers_0_125):
    #input: list of integers exactly as received from K-DUCER response to read_input_registers(reg_addr=0, reg_nb=125)
    #returns: string containing screwdriving results
    #this is necessary because pyModbusTCP assumes all responses are integers, but the K-DUCER response is an ASCII coded string
    #must convert each integer to two bytes in big endian (modbus uses big endian), then decode as ASCII
    if output_from_function_read_input_registers_0_125 == None or type(output_from_function_read_input_registers_0_125) != type(['']) or len(output_from_function_read_input_registers_0_125) == 0:
        print('Input must be the list returned by the pyModbusTCP function: read_input_registers(reg_addr=0, reg_nb=125). Input was: {}'.format(output_from_function_read_input_registers_0_125))
        return ''
    return (b''.join([i.to_bytes(2, byteorder='big') for i in output_from_function_read_input_registers_0_125]).decode(encoding='ascii',errors='ignore')).replace('\00','').replace('\r\n','\n')

def get_graph_data(MBclient):
    time.sleep(0.25)
    TT_graph_data = MBclient.read_input_registers(reg_addr= address_offset+153, reg_nb= 71)
    TA_graph_data = MBclient.read_input_registers(reg_addr= address_offset+224, reg_nb= 71)
    TT_graph_data.extend(TA_graph_data)
    return TT_graph_data
    
def graph_results(graph_data, print_string):
    #print(graph_data)

    TT_sample_time = graph_data[0]
    TT_data_y = np.trim_zeros(graph_data[1:70], 'b')

    TA_sample_time = graph_data[71]
    TA_data_y = np.trim_zeros(graph_data[72:], 'b')

    #Screw NOK results can have unequal length of TA/TT results
    min_length = min([len(TT_data_y), len(TA_data_y)])
    TT_data_y = TT_data_y[:min_length]
    TA_data_y = TA_data_y[:min_length]

    TT_data_x = range(0,len(TT_data_y)*TT_sample_time,TT_sample_time)
    TA_data_x = range(0,len(TA_data_y)*TA_sample_time,TA_sample_time)

    plt.subplot(131, title='Torque vs Time')
    plt.plot(TT_data_x, TT_data_y, marker='.')
    plt.grid()
    plt.ylabel('Torque, cNm')
    plt.xlabel('Time, ms')

    plt.subplot(132, title='Torque vs Angle')
    plt.plot(TA_data_y, TT_data_y, marker='.')
    plt.grid()
    plt.ylabel('Torque, cNm')
    plt.xlabel('Angle, degrees')

    plt.subplot(133, title='Angle vs Time')
    plt.plot(TA_data_x, TA_data_y, marker='.')
    plt.grid()
    plt.ylabel('Angle, degrees')
    plt.xlabel('Time, ms')

    for i in range(10,0,-1):
        plt.suptitle('K-Ducer Torque/Angle data from last screwdriving. Obtained via MODBUS TCP\n\n{}\nNew MODBUS remote-controlled screwdriving in {} seconds...'.format(print_string,i))
        plt.pause(1)
    
    plt.clf()

if __name__ == "__main__":
    for i in range(10,0,-1):
        print('******WARNING******* screwdriver motor will start running in {} seconds. Press CTRL+c to abort script.'.format(i))
        time.sleep(1)

    exampleMBClient = ModbusClient(host=kdu_ip_address, port=502, auto_open=True, auto_close=False, timeout=3) #instantiate a MODBUS client. auto_close must be False for REMOTE_LEVER control
    exampleMBClient.debug(False) #use True if you want the script to print all the MODBUS packets and debug informantion on the console
    
    plt.figure(clear=True, figsize=[14,6], dpi = 100, tight_layout=True)

    #run screwdriver. If Screw OK, repeat with program #1. If Screw NOK, repeat with program #2. 
    next_program = 1
    while True:
        change_current_program(MBclient=exampleMBClient, program_number=next_program)
        final_screwdriving_state = run_motor_until_it_stops(MBclient=exampleMBClient)
        screwdriving_results_string = convert_kducer_printstring(exampleMBClient.read_input_registers(reg_addr= address_offset+1, reg_nb= 125))
        print(screwdriving_results_string)
        if (final_screwdriving_state == 13) or (final_screwdriving_state == 14): # 13 = Screw OK (torque control mode), 14 = Angle OK (angle control mode). Refer to K-Ducer MODBUS map.
            next_program = 1
            print('Screw OK. Selecting program 1')
        else:
            next_program = 2
            print('Screw N-OK. Selecting program 2')
        graph_results(get_graph_data(exampleMBClient), screwdriving_results_string)
